﻿
' =======================================================================================
'  Class Filter_MultiPole
' =======================================================================================
'  This is an IIR filter, like multiple RC cells
'  The tau is correct so the fCut is alwaiys at -3dB point, also using many poles
' =======================================================================================

Imports System.Math
Imports System.Collections.Generic

Public Class Filter_MultiPole

    Private Const MAX_POLES As Integer = 100
    '
    Private sw As System.Diagnostics.Stopwatch = New System.Diagnostics.Stopwatch
    Private lastMS As Double         ' last time in milliseconds
    '
    Private fCut3dB As Double
    Private nPoles As Integer
    Private highPass As Boolean
    Private tauMS As Double
    Private fC As Double
    Private sampleTime As Double
    Private tauSamples As Double
    Private ampFactor As Double
    '
    Private buffer(MAX_POLES - 1) As Double
    Private bufferHiPass(MAX_POLES - 1) As Double

    Private Sub CalculateTau()
        If Me.highPass Then
            Me.fC = Me.fCut3dB * Sqrt(Pow(2, (1 / Me.nPoles)) - 1)
        Else
            Me.fC = Me.fCut3dB / Sqrt(Pow(2, (1 / Me.nPoles)) - 1)
        End If
        Me.tauMS = 1000 / (2 * (PI * Me.fC))
    End Sub

    Private Sub SetDefaultValues()
        Me.fCut3dB = 100
        Me.nPoles = 1
        Me.highPass = False
        Me.CalculateTau()
    End Sub

    Friend Sub New()
        MyBase.New()
        sw.Start()
        buffer(0) = 0
        bufferHiPass(0) = 0
        lastMS = sw.Elapsed.TotalMilliseconds
        Me.SetDefaultValues()
    End Sub

    Friend Sub SetFrequency(ByVal _fCut3dB As Double)
        If Me.fCut3dB = _fCut3dB Then Return
        Me.fCut3dB = _fCut3dB
        Me.CalculateTau()
    End Sub

    Friend Sub SetNpoles(ByVal _nPoles As Integer)
        If _nPoles > MAX_POLES Then Error 98
        If Me.nPoles = _nPoles Then Return
        Me.nPoles = _nPoles
        Me.CalculateTau()
    End Sub

    Friend Sub SetHiPass(ByVal _hiPass As Boolean)
        If Me.highPass = _hiPass Then Return
        Me.highPass = _hiPass
        Me.CalculateTau()
    End Sub

    Friend Function Exec(ByVal inValue As Double) As Double
        ' -----------------------------------------------------------------
        Dim time As Double = sw.Elapsed.TotalMilliseconds
        Me.sampleTime = time - lastMS
        lastMS = time
        ' -----------------------------------------------------------------
        If (Me.nPoles = 0) Then Return inValue
        ' -----------------------------------------------------------------
        Me.tauSamples = Me.tauMS / Me.sampleTime
        Me.ampFactor = Exp(-1 / Me.tauSamples)
        ' -----------------------------------------------------------------
        If Me.highPass Then
            bufferHiPass(Me.nPoles) = inValue
            For i As Int32 = Me.nPoles - 1 To 0 Step -1
                buffer(i) = (1 - Me.ampFactor) * bufferHiPass(i + 1) + Me.ampFactor * buffer(i)
                bufferHiPass(i) = bufferHiPass(i + 1) - buffer(i)
            Next
        Else
            buffer(Me.nPoles) = inValue
            For i As Int32 = Me.nPoles - 1 To 0 Step -1
                buffer(i) = (1 - Me.ampFactor) * buffer(i + 1) + Me.ampFactor * buffer(i)
            Next
        End If
        ' -----------------------------------------------------------------
        If Me.highPass Then
            Return bufferHiPass(0)
        Else
            Return buffer(0)
        End If
        ' -----------------------------------------------------------------
    End Function

End Class


Module Filters

    Private Filters As Dictionary(Of Integer, Filter_MultiPole) = New Dictionary(Of Integer, Filter_MultiPole)

    Friend Sub ResetAllFilters()
        Filters.Clear()
    End Sub

    Friend Function ExecFilter(ByVal v As Double, _
                               ByVal freq As Double, _
                               ByVal npoles As Double, _
                               ByVal HiPass As Boolean, _
                               ByVal identifier As Int32) As Double
        ' -----------------------------------------------
        If Not Filters.ContainsKey(identifier) Then
            Filters.Add(identifier, New Filter_MultiPole)
        End If
        ' -----------------------------------------------
        With Filters.Item(identifier)
            .SetFrequency(freq)
            .SetNpoles(CInt(npoles))
            .SetHiPass(HiPass)
            Return .Exec(v)
        End With
    End Function

End Module

